#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import multiprocessing
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Optional, IO, Dict

from .packageparser import PackageParser
from . import ADBClient
from . import ActivityManager
from . import PackageManager
from . import adb_connection
from . import propertyparser

__all__ = ["ADBDevice"]


def log():
    from . import _logger
    logger = _logger.get_logger(__name__)
    return logger


class ADBDevice(object):
    def __init__(self, client: ADBClient):
        self._client = client
        self._name: Optional[str] = None
        self._am = ActivityManager(self._client)
        self._pm = PackageManager(self._client)
        self._executor = ThreadPoolExecutor(max_workers=multiprocessing.cpu_count())

    @property
    def client(self) -> ADBClient:
        return self._client

    @property
    def am(self) -> ActivityManager:
        return self._am

    @property
    def pm(self) -> PackageManager:
        return self._pm

    @property
    def name(self) -> Optional[str]:
        if self._name is None:
            self._name = self.client.getprop("ro.build.product")
        return self._name

    @property
    def api_level(self):
        return int(self.client.getprop("ro.build.version.sdk"))

    @property
    def android_release_version(self):
        return self.client.getprop("ro.build.version.release")

    @property
    def build_prop(self) -> propertyparser.PropertyParser:
        return self._read_build_prop()

    def quit(self):
        log().info("Quitting...")
        self._executor.shutdown()
        self._client = None

    def write_screencap(self, fp: IO) -> bool:
        log().verbose("write screencap")
        return self.client.execute("exec-out screencap -p", stdout=fp)

    def save_screencap(self, output: str):
        log().verbose(f"save screencap to {output}")
        return self.client.shell("screencap -p %s" % output)

    def screenrecord(
            self,
            file: str,
            bugreport: bool = False,
            bitrate: int = 8000000,
            timelimit: int = 0,
            **kwargs,
    ) -> bool:
        log().verbose(f"screenrecord to {file}")

        args = ["--bit-rate", str(bitrate)]
        if bugreport:
            args.append("--bugreport")
        if timelimit > 0:
            args.extend(("--time-limit", str(timelimit)))
        args.append(file)

        return self.client.shell(
            "screenrecord", **adb_connection.extends_extra_arguments(*args, **kwargs)
        )

    """ -----------------------------------------------------------------------"""
    """ Events """
    """ -----------------------------------------------------------------------"""

    def async_send_key(self, key: int):
        return self._executor.submit(self.client.send_key, key)

    def async_send_text(self, char: str):
        assert type(char) is str
        return self._executor.submit(self.client.send_text, char)

    """ -----------------------------------------------------------------------"""
    """ Activities """
    """ -----------------------------------------------------------------------"""

    def open_settings(self):
        log().verbose("Opening Settings Activity...")
        return self.client.shell(
            "am start -a android.intent.action.MAIN -n com.android.settings/.Settings"
        )

    def open_bluetooth_settings(self):
        log().verbose("Opening Blueooth Settings Activity...")
        return self.client.shell("am start -a android.settings.BLUETOOTH_SETTINGS")

    def open_monitor_client(self):
        log().verbose("Opening Monitor Client...")
        return self.client.shell(
            "am start -a android.intent.action.MAIN -n com.iwedia.stbmclient/com.example.stbmclient.STBMonitorClientActivity"
        )

    """ DumpSys Wrapper """

    def get_services_info(self) -> Optional[Dict]:
        result = self.client.dumpsys("package r service")
        if result.is_ok():
            p = PackageParser(result.output())
            if 'Service Resolver Table:' in p:
                if 'Non-Data Actions:' in p['Service Resolver Table:']:
                    return p['Service Resolver Table:']['Non-Data Actions:']
        return None

    def get_receivers_info(self) -> Optional[Dict]:
        """
        Executes `dumpsys package r receiver` on the connected device and return a table
        with all the registered receivers
        :return:
        """
        result = self.client.dumpsys("package r receiver")
        if result.is_ok():
            p = PackageParser(result.output())
            if 'Receiver Resolver Table:' in p:
                if 'Non-Data Actions:' in p['Receiver Resolver Table:']:
                    return p['Receiver Resolver Table:']['Non-Data Actions:']
        return None

    """ -----------------------------------------------------------------------"""
    """ Private Methods """
    """ -----------------------------------------------------------------------"""

    def _read_build_prop(self) -> propertyparser.PropertyParser:
        parser = propertyparser.PropertyParser()
        if self.client.root():
            properties = self.client.cat("/system/build.prop")
            if properties:
                return parser.read_string(properties)
        return parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Optional, Any
from enum import Enum

__all__ = ["ADBEvent", "ADBConnectionEvent"]


class ADBEvent(object):
    def __init__(self, event_type: ADBConnectionEvent, event_data: Optional[Any]):
        self._event_type = event_type
        self._event_data = event_data

    @property
    def event_type(self):
        return self._event_type

    @property
    def event_data(self):
        return self._event_data

    def __str__(self):
        return f"ADBEvent(type={self._event_type}, data={self._event_data})"


class ADBConnectionEvent(Enum):
    Reboot = "reboot"
    Connect = "connect"
    Disconnect = "disconnect"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import logging

import coloredlogs
import verboselogs

__all__ = ["get_logger", "set_logging_factory"]

logging_factory_func = None

loggers = dict()


def get_logger(name: str) -> verboselogs.VerboseLogger:
    global logging_factory_func

    if logging_factory_func:
        # noinspection PyCallingNonCallable
        return logging_factory_func(name)

    global loggers
    if name not in loggers:
        logger = verboselogs.VerboseLogger(name)
    else:
        logger = loggers[name]
    return logger


def set_logging_factory(func):
    global logging_factory_func
    logging_factory_func = func



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from . import Intent
from . import ADBClient

__all__ = ["ActivityManager"]


def log():
    from . import _logger
    return _logger.get_logger(__name__)


class ActivityManager(object):
    def __init__(self, client: ADBClient):
        self._client = client

    @property
    def client(self):
        return self._client

    def broadcast(self, intent: Intent):
        arguments = intent.build()
        log().verbose(f"broadcast with arguments {arguments}")
        return self.client.shell(f"am broadcast {arguments}")

    def startservice(self, intent: Intent):
        arguments = intent.build()
        log().verbose(f"startservice with arguments {arguments}")
        return self.client.shell(f"am startservice {arguments}")

    def start(self, intent: Intent):
        arguments = intent.build()
        log().verbose(f"start with arguments {arguments}")
        return self.client.shell(f"am start {arguments}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from typing import Dict

__all__ = ["Package"]


class Package(object):
    # noinspection PyShadowingBuiltins
    def __init__(self, input: Dict[str, str]):
        self._name = input["package"]
        self._apk = input.get("apk", None)
        self._installer = input.get("installer", None)
        self._uuid = input.get("uuid", None)

    @property
    def name(self):
        return self._name

    @property
    def apk(self):
        return self._apk

    @property
    def installer(self):
        return self._installer

    @property
    def uuid(self):
        return self._uuid

    def is_system(self):
        return self._apk.startswith("/system/") if self._apk else False

    def __repr__(self):
        return f"Package<id: {id(self)}, name: {self.name}>"

    def __str__(self):
        return f"Package({self.name})"

    def __eq__(self, other):
        if isinstance(other, (Package,)):
            return self.name == other.name
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Android Debug Bridge version 1.0.41
Version 30.0.4-6686687
Installed as /Users/alessandro/Library/Android/sdk/platform-tools/adb

global options:
 -a         listen on all network interfaces, not just localhost
 -d         use USB device (error if multiple devices connected)
 -e         use TCP/IP device (error if multiple TCP/IP devices available)
 -s SERIAL  use device with given serial (overrides $ANDROID_SERIAL)
 -t ID      use device with given transport id
 -H         name of adb server host [default=localhost]
 -P         port of adb server [default=5037]
 -L SOCKET  listen on given socket for adb server [default=tcp:localhost:5037]

general commands:
 devices [-l]             list connected devices (-l for long output)
 help                     show this help message
 version                  show version num

networking:
 connect HOST[:PORT]      connect to a device via TCP/IP [default port=5555]
 disconnect [HOST[:PORT]]
     disconnect from given TCP/IP device [default port=5555], or all
 pair HOST[:PORT] [PAIRING CODE]
     pair with a device for secure TCP/IP communication
 forward --list           list all forward socket connections
 forward [--no-rebind] LOCAL REMOTE
     forward socket connection using:
       tcp:<port> (<local> may be "tcp:0" to pick any open port)
       localabstract:<unix domain socket name>
       localreserved:<unix domain socket name>
       localfilesystem:<unix domain socket name>
       dev:<character device name>
       jdwp:<process pid> (remote only)
       acceptfd:<fd> (listen only)
 forward --remove LOCAL   remove specific forward socket connection
 forward --remove-all     remove all forward socket connections
 ppp TTY [PARAMETER...]   run PPP over USB
 reverse --list           list all reverse socket connections from device
 reverse [--no-rebind] REMOTE LOCAL
     reverse socket connection using:
       tcp:<port> (<remote> may be "tcp:0" to pick any open port)
       localabstract:<unix domain socket name>
       localreserved:<unix domain socket name>
       localfilesystem:<unix domain socket name>
 reverse --remove REMOTE  remove specific reverse socket connection
 reverse --remove-all     remove all reverse socket connections from device
 mdns check               check if mdns discovery is available
 mdns services            list all discovered services

file transfer:
 push [--sync] [-z ALGORITHM] [-Z] LOCAL... REMOTE
     copy local files/directories to device
     --sync: only push files that are newer on the host than the device
     -n: dry run: push files to device without storing to the filesystem
     -z: enable compression with a specified algorithm (any, none, brotli)
     -Z: disable compression
 pull [-a] [-z ALGORITHM] [-Z] REMOTE... LOCAL
     copy files/dirs from device
     -a: preserve file timestamp and mode
     -z: enable compression with a specified algorithm (any, none, brotli)
     -Z: disable compression
 sync [-l] [-z ALGORITHM] [-Z] [all|data|odm|oem|product|system|system_ext|vendor]
     sync a local build from $ANDROID_PRODUCT_OUT to the device (default all)
     -n: dry run: push files to device without storing to the filesystem
     -l: list files that would be copied, but don't copy them
     -z: enable compression with a specified algorithm (any, none, brotli)
     -Z: disable compression

shell:
 shell [-e ESCAPE] [-n] [-Tt] [-x] [COMMAND...]
     run remote shell command (interactive shell if no command given)
     -e: choose escape character, or "none"; default '~'
     -n: don't read from stdin
     -T: disable pty allocation
     -t: allocate a pty if on a tty (-tt: force pty allocation)
     -x: disable remote exit codes and stdout/stderr separation
 emu COMMAND              run emulator console command

app installation (see also `adb shell cmd package help`):
 install [-lrtsdg] [--instant] PACKAGE
     push a single package to the device and install it
 install-multiple [-lrtsdpg] [--instant] PACKAGE...
     push multiple APKs to the device for a single package and install them
 install-multi-package [-lrtsdpg] [--instant] PACKAGE...
     push one or more packages to the device and install them atomically
     -r: replace existing application
     -t: allow test packages
     -d: allow version code downgrade (debuggable packages only)
     -p: partial application install (install-multiple only)
     -g: grant all runtime permissions
     --abi ABI: override platform's default ABI
     --instant: cause the app to be installed as an ephemeral install app
     --no-streaming: always push APK to device and invoke Package Manager as separate steps
     --streaming: force streaming APK directly into Package Manager
     --fastdeploy: use fast deploy
     --no-fastdeploy: prevent use of fast deploy
     --force-agent: force update of deployment agent when using fast deploy
     --date-check-agent: update deployment agent when local version is newer and using fast deploy
     --version-check-agent: update deployment agent when local version has different version code and using fast deploy
     --local-agent: locate agent files from local source build (instead of SDK location)
     (See also `adb shell pm help` for more options.)
 uninstall [-k] PACKAGE
     remove this app package from the device
     '-k': keep the data and cache directories

debugging:
 bugreport [PATH]
     write bugreport to given PATH [default=bugreport.zip];
     if PATH is a directory, the bug report is saved in that directory.
     devices that don't support zipped bug reports output to stdout.
 jdwp                     list pids of processes hosting a JDWP transport
 logcat                   show device log (logcat --help for more)

security:
 disable-verity           disable dm-verity checking on userdebug builds
 enable-verity            re-enable dm-verity checking on userdebug builds
 keygen FILE
     generate adb public/private key; private key stored in FILE,

scripting:
 wait-for[-TRANSPORT]-STATE...
     wait for device to be in a given state
     STATE: device, recovery, rescue, sideload, bootloader, or disconnect
     TRANSPORT: usb, local, or any [default=any]
 get-state                print offline | bootloader | device
 get-serialno             print <serial-number>
 get-devpath              print <device-path>
 remount [-R]
      remount partitions read-write. if a reboot is required, -R will
      will automatically reboot the device.
 reboot [bootloader|recovery|sideload|sideload-auto-reboot]
     reboot the device; defaults to booting system image but
     supports bootloader and recovery too. sideload reboots
     into recovery and automatically starts sideload mode,
     sideload-auto-reboot is the same but reboots after sideloading.
 sideload OTAPACKAGE      sideload the given full OTA package
 root                     restart adbd with root permissions
 unroot                   restart adbd without root permissions
 usb                      restart adbd listening on USB
 tcpip PORT               restart adbd listening on TCP on PORT

internal debugging:
 start-server             ensure that there is a server running
 kill-server              kill the server if it is running
 reconnect                kick connection from host side to force reconnect
 reconnect device         kick connection from device side to force reconnect
 reconnect offline        reset offline/unauthorized devices to force reconnect

environment variables:
 $ADB_TRACE
     comma-separated list of debug info to log:
     all,adb,sockets,packets,rwx,usb,sync,sysdeps,transport,jdwp
 $ADB_VENDOR_KEYS         colon-separated list of keys (files or directories)
 $ANDROID_SERIAL          serial number to connect to (see -s)
 $ANDROID_LOG_TAGS        tags to be used by logcat (see logcat --help)
 $ADB_LOCAL_TRANSPORT_MAX_PORT max emulator scan port (default 5585, 16 emus)
 $ADB_MDNS_AUTO_CONNECT   comma-separated list of mdns services to allow auto-connect (default adb-tls-connect)

"""

from __future__ import annotations

import re
import shutil
import subprocess
import time
from typing import List, Optional, Dict, Tuple, Any

# noinspection PyPackageRequirements
import zope.event

from .. import events
from .. import constants

__all__ = [
    "ADBCommandResult",
    "Device",
    "bugreport",
    "busybox",
    "capture_output",
    "connect",
    "devices",
    "disconnect",
    "disconnect_all",
    "execute",
    "get_adb_path",
    "is_connected",
    "is_root",
    "mdns_check",
    "mdns_services",
    "pull",
    "push",
    "reboot",
    "reconnect",
    "reconnect_device",
    "reconnect_offline",
    "remount",
    "remount_as",
    "root",
    "set_adb_path",
    "shell",
    "unroot",
    "version",
    "wait_for_device",
    "which",
    "extends_extra_arguments",
    "expand_extra_arguments",
]

adb_path = None


class Device(object):
    """
    Represententation of a device, as listed from 'adb devices -l'
    """

    def __init__(self, identifier: str, attrs: Dict[str, str]):
        super(Device, self).__init__()
        self._identifier = identifier
        self._attrs = attrs

    @property
    def identifier(self):
        return self._identifier

    @property
    def transport_id(self):
        return self._attrs["transport_id"] if "transport_id" in self._attrs else None

    def is_usb(self) -> bool:
        """
        Return true if the attached device is connected through USB
        :return:
        """
        return "usb" in self._attrs

    def is_emulator(self) -> bool:
        """
        Return true if the attached device is an emulator
        :return:
        """
        return self._identifier.startswith("emulator-")

    def is_wifi(self) -> bool:
        """
        Return true if the attached device is connected through wi-fi
        :return:
        """
        if not self.is_usb() and not self.is_emulator():
            return re.match(Device.PATTERN_IP, self._identifier) is not None
        return False

    def __getattribute__(self, item):
        return object.__getattribute__(self, item)

    def __getattr__(self, item):
        if item in self._attrs:
            return self._attrs[item]
        return None

    def __str__(self):
        string = f"Device(id={self._identifier},"
        for key, value in self._attrs.items():
            string += f"{key}={value},"
        string += ")"
        return string

    PATTERN_LINE = re.compile("^([^\\s]+)\\s+device\\s+(.*)$")
    PATTERN_ATTR = re.compile("(\\w+:\\w+)")
    PATTERN_IP = re.compile(
        "^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]):[0-9]+$"
    )

    @staticmethod
    def parse(string: str) -> Optional[Device]:
        match1 = re.match(Device.PATTERN_LINE, string)
        if match1:
            device_id = match1.group(1)
            attrbutes = match1.group(2)
            device_attr = dict(
                map(lambda x: x.split(":"), re.findall(Device.PATTERN_ATTR, attrbutes))
            )
            return Device(device_id, device_attr)
        else:
            log().warning(f"Failed to parse device string {string}")
            return None


class ADBCommandResult(object):
    RESULT_OK = 0
    RESULT_ERROR = 1

    def __init__(
            self,
            result: Tuple[Optional[bytes], Optional[bytes]] = (None, None),
            code: int = RESULT_OK,
    ):
        super(ADBCommandResult, self).__init__()
        self._stdout = result[0] if result[0] else None
        self._stderr = (
            result[1].decode("utf-8").strip() if result[1] is not None else None
        )
        self._code = code

    def __iter__(self):
        return (self._code, self.output(), self.error()).__iter__()

    def __str__(self):
        return f"Result(code={self._code}, output={self._stdout}, error={self._stderr})"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def code(self):
        """
        Error Code returned. if '0' it means no error
        :return:
        """
        return self._code

    def output(self, raw: bool = False):
        if self._stdout and not raw:
            return self._stdout.decode("utf-8").strip()
        return self._stdout

    def error(self):
        return self._stderr

    def is_ok(self):
        return self.code == ADBCommandResult.RESULT_OK

    def is_error(self):
        return not self.is_ok()

    @staticmethod
    def from_error(code: int = RESULT_ERROR):
        return ADBCommandResult(result=(None, None), code=code)


def log():
    from .. import _logger

    logger = _logger.get_logger(__name__)
    return logger


def get_adb_path() -> str:
    """
    Attempts to retrieve the current adb path from the system environment
    :return:
    """
    global adb_path
    if adb_path is None:
        adb_path = shutil.which("adb")
        log().debug(f"Found adb: {adb_path}")

    assert adb_path
    return adb_path


def set_adb_path(path: str):
    """
    Set the global adb path
    :param path: absolute adb path
    """
    global adb_path
    adb_path = path


def busybox(command: str, ip: Optional[str] = None, **kwargs):
    """
    Executes a shell command using 'busybox'. If busybox in not installed on the device it will fail.\n
    See https://busybox.net/
    :param command:
    :param ip:
    :param kwargs:
    :return:
    """
    return shell(f"{constants.BUSYBOX} {command}", ip, **kwargs)


def which(command: str, ip: Optional[str] = None) -> Optional[str]:
    """
    Execute 'which' command on the specified device and return the result
    :param command:     command to test with 'which'
    :param ip:          device ip
    :return:            the command path on the device, if found
    """
    which_cmd = f"{constants.WHICH}"
    result = shell(f"{which_cmd} {command}", ip=ip)
    if result.code == ADBCommandResult.RESULT_OK and result.output() and len(result.output()) > 0:
        return result.output()
    return None


def execute(command: str, ip: Optional[str] = None, **kwargs) -> bool:
    """
    Execute an adb command on the given device, if connected
    :param command: command to execute
    :param ip:      device identifier
    :param kwargs:
    :return:
    """
    ip_arg = f"-s {ip} " if ip else ""
    adb = get_adb_path()
    command_line = f"{ip_arg}{command} {get_extra_arguments(**kwargs)}"
    command_full = f"{adb} {command_line}".strip()
    command_log = f"adb {command_line}".strip()
    output = subprocess.PIPE if "stdout" not in kwargs else kwargs["stdout"]

    try:
        log().debug(f"Executing `{command_log}`")
        out = subprocess.Popen(
            command_full.split(), stderr=subprocess.STDOUT, stdout=output
        )
        result = out.communicate()
        if not command == "get-state":
            result_str = result[0].decode("utf-8").strip() if result[0] else "None"
            log().spam(f"Result(code={out.returncode}, result={result_str})")
        return out.returncode == ADBCommandResult.RESULT_OK
    except subprocess.CalledProcessError:
        return False


def capture_output(
        command: str, ip: Optional[str] = None, **kwargs
) -> ADBCommandResult:
    """
    Execute an adb command on the given device and return the result
    :param command:  command to execute
    :param ip:       device id
    :param kwargs:   if contains the 'args' key, its value is passed as arguments to the input command
    :return:
    """
    ip_arg = f"-s {ip} " if ip else ""
    adb = get_adb_path()
    command_line = f"{ip_arg}{command} {get_extra_arguments(**kwargs)}"
    command_full = f"{adb} {command_line}".strip()
    command_log = f"adb {command_line}".strip()
    try:
        if kwargs.get("log", True):
            log().debug(f"Executing `{command_log}`")
        out = subprocess.Popen(
            command_full.split(),
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            shell=False,
        )
        result = out.communicate()
        return ADBCommandResult(result, out.returncode)
    except subprocess.CalledProcessError as e:
        log().warning(e)
        return ADBCommandResult(code=ADBCommandResult.RESULT_ERROR)


def shell(command: str, ip: Optional[str] = None, **kwargs) -> ADBCommandResult:
    """
    Execute a command in the adb shell
    :param command:         shell command to execute
    :param ip:              device ip
    :param kwargs:          if contains the 'args' key, its value is passed as arguments to the input command
    :return:
    """
    ip_arg = f"-s {ip} " if ip else ""
    adb = get_adb_path()
    command_line = f"{ip_arg} shell {command} {get_extra_arguments(**kwargs)}"
    command_full = f"{adb} {command_line}".strip()
    command_log = f"adb {command_line}".strip()
    log().debug(f"Executing `{command_log}`")

    out = subprocess.Popen(
        command_full.split(),
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        shell=False,
    )

    result = out.communicate()
    code = out.returncode
    return ADBCommandResult(result, code)


def wait_for_device(ip: Optional[str] = None) -> bool:
    """
    Execute and return the result of 'adb wait-for-device'
    :param ip:              device ip
    :return:                true if device connects
    """
    # if is_connected(ip=ip):
    #     return True

    execute(
        "wait-for-device shell 'while [[ -z $(getprop sys.boot_completed) ]]; do sleep 1; done; input keyevent 143'",
        ip=ip,
    )

    return is_connected(ip=ip)


def reconnect(ip: Optional[str] = None) -> bool:
    """
    kick connection from host side to force reconnect

    :param ip:  device id
    :return:    true on success
    """
    return execute("reconnect", ip=ip)


def reconnect_device(ip: Optional[str] = None) -> bool:
    """
    kick connection from device side to force reconnect
    :param ip:  device id
    :return:    true on success
    """
    return execute("reconnect device", ip=ip)


def reconnect_offline(ip: Optional[str] = None) -> bool:
    """
    reset offline/unauthorized devices to force reconnect

    :param ip:  device id
    :return:    true on success
    """
    return execute("reconnect offline", ip=ip)


def root(ip: Optional[str] = None) -> bool:
    """
    Restart adb connection as root
    :param ip:              device ip
    :return:                true if root succeded
    """
    # if is_root(ip=ip):
    #     return True
    # execute("root", ip=ip)
    # execute("reconnect", ip=ip)
    # wait_for_device(ip=ip)
    # return is_root(ip=ip)
    return execute("root", ip=ip)


# noinspection SpellCheckingInspection
def unroot(ip: Optional[str] = None) -> bool:
    """
    Restart adb as unroot
    :param ip:              device ip
    :return:                True on success
    """
    # if not is_root(ip=ip):
    #     return True
    #
    # execute("unroot", ip=ip)
    # reconnect(ip=ip)
    # wait_for_device(ip=ip)
    # return not is_root(ip=ip)
    return execute("unroot", ip=ip)


def is_root(ip: Optional[str] = None) -> bool:
    """
    Test if adb connection is running as root
    :param ip:              device ip
    :return:                True if adb is running as root
    """
    result = shell("whoami", ip=ip)
    if result.code == ADBCommandResult.RESULT_OK:
        log().spam(f"whoami: {result.output()}")
        return result.output() == "root"
    return False


def devices(**kwargs) -> List[str]:
    """
    Return a list of attached devices
    :param kwargs: extra arguments
    :return:
    """
    result = capture_output(command="devices", **kwargs)
    if result.code == ADBCommandResult.RESULT_OK:
        return list(map(lambda x: x.split("\t")[0], result.output().split("\n")[1:]))
    else:
        return []
    # if result.code == ADBCommandResult.RESULT_OK and result.result:
    #     attached_devices = list(map(lambda x: x.split("\t")[0], result.result.split("\n")[1:]))
    #     result = list(filter(lambda y: y is not None, map(lambda x: Device.parse(x), attached_devices)))
    #     return result
    # else:
    #     return []


def is_connected(ip: Optional[str] = None) -> bool:
    """
    Returns true if the given device is connected
    :param ip:  device ip
    :return: true if connected
    """
    result = capture_output(command="get-state", ip=ip, log=False)
    # log().spam(result)
    return result.code == ADBCommandResult.RESULT_OK and result.output() == "device"


def connect(ip: Optional[str]) -> bool:
    """
    Attempts to connect to the device with the given ip address
    :param ip: device ip
    :return: true if connection succeed
    """
    log().debug(f"Connecting to {ip}..")

    if is_connected(ip=ip):
        log().info("Already connected...")
        return True

    try:
        out = subprocess.Popen(
            f"{get_adb_path()} connect {ip}",
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            shell=True,
        )
        result = out.communicate()  # noqa: F841
        time.sleep(50 / 1000)
        if out.returncode == 0 and is_connected(ip):
            zope.event.notify(events.ADBEvent(events.ADBConnectionEvent.Connect, ip))
            return True
        else:
            return False
    except subprocess.CalledProcessError:
        return False


def disconnect_all() -> bool:
    """
    Disconnect all connected devices
    :return: true if successfully disconnected from all devices
    """
    log().verbose("Disconnecting from everything..")
    if execute("disconnect"):
        zope.event.notify(events.ADBEvent(events.ADBConnectionEvent.Disconnect, None))
        return True
    return False


def disconnect(ip: Optional[str]) -> bool:
    """
    Disconnect from the given device ip
    :param ip: device ip
    :return: true if successfully disconnected
    """
    log().verbose(f"Disconnecting from {ip}..")
    if execute("disconnect", ip=ip):
        zope.event.notify(events.ADBEvent(events.ADBConnectionEvent.Disconnect, ip))
        return True
    return False


def reboot(ip: Optional[str] = None, mode: Optional[str] = None) -> bool:
    """
    Reboot the device
    :param ip:              device ip
    :param mode:            reboot type
    :return:                True on success
    """
    command = "reboot {}".format(mode if mode else "")
    if execute(command, ip=ip):
        zope.event.notify(events.ADBEvent(events.ADBConnectionEvent.Reboot, ip))
        return True
    return False


def remount(ip: Optional[str] = None) -> bool:
    """
    Remount the file system if root is permitted. Requires root
    :param ip:              device ip
    :return:                True on success
    """
    return execute("remount", ip=ip)


def remount_as(
        ip: Optional[str] = None, writeable: bool = False, folder: str = "/system"
) -> bool:
    """
    Mount/Remount file-system. Requires root
    :param folder:          folder to mount
    :param writeable:       mount as writeable or readable-only
    :param ip:              device ip
    :rtype:                 true on success
    """
    if writeable:
        return (
                shell(f"mount -o rw,remount {folder}", ip=ip).code
                == ADBCommandResult.RESULT_OK
        )
    else:
        return (
                shell(f"mount -o ro,remount {folder}", ip=ip).code
                == ADBCommandResult.RESULT_OK
        )


def version() -> Optional[str]:
    """
    Fetch and return the current adb version
    :return: the output as returned by 'adb version'
    """
    result = capture_output(command="version")
    if result.code == ADBCommandResult.RESULT_OK:
        return result.output()
    return None


def mdns_services() -> ADBCommandResult:
    """
    Return the result of the command 'adb mdns services'
    :return:
    """
    return capture_output(command="mdns services")


def mdns_check() -> ADBCommandResult:
    """
    Return the result of the command 'adb mdns check'
    :return:
    """
    return capture_output(command="mdns check")


def bugreport(dest: Optional[str] = None, ip: Optional[str] = None) -> bool:
    """
    Execute and return the result of the command 'adb bugreport'
    :param dest:    optional target local folder/filename for the bugreport
    :param ip:      optional device ip
    :return:        true on success
    """
    return execute("bugreport", ip=ip, args=(dest if dest else "",))


def push(ip: Optional[str], src: str, dst: str, **kwargs) -> bool:
    return execute(command="push", ip=ip, **extends_extra_arguments(src, dst, **kwargs))


def pull(ip: Optional[str], src: str, dst: str, **kwargs):
    return execute(command="pull", ip=ip, **extends_extra_arguments(src, dst, **kwargs))


""" Utilities """


def get_extra_arguments(**kwargs) -> str:
    _check_extra_arguments_type(**kwargs)
    if "args" in kwargs:
        return " ".join(filter(lambda x: x is not None, kwargs["args"]))
    return ""


def expand_extra_arguments(**kwargs) -> List[str]:
    _check_extra_arguments_type(**kwargs)
    if "args" in kwargs:
        return list(filter(lambda x: x is not None, kwargs["args"]))
    return []


def extends_extra_arguments(*args, **kwargs) -> Dict[Any, Any]:
    _check_extra_arguments_type(**kwargs)
    newlist = expand_extra_arguments(**kwargs)
    newlist.extend(args)
    kwargs["args"] = newlist
    return kwargs


def _check_extra_arguments_type(**kwargs):
    if "args" in kwargs:
        if not isinstance(kwargs["args"], (tuple, list)):
            raise RuntimeError(
                f'`args` must be either a tuple or a list, got {type(kwargs["args"])} instead'
            )



from ._version import __version__
from .types import *
from .intent import Intent
from .keycodes import KeyCodes, CPlusPlusKeyCodes
from . import adb_connection
from .adb_client import ADBClient
from .am import ActivityManager
from .pm import PackageManager
from .adb_device import ADBDevice
from ._logger import set_logging_factory

__author__ = "Alessandro Crugnola"

__doc__ = f"""
Python ADB Client
version: {__version__}
author: {__author__}
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import optional
from pathlib import Path
from typing import Dict, Union, Optional

from . import adb_connection
from . import constants
from .adb_connection import ADBCommandResult

__all__ = ["ADBClient"]


def log():
    from . import _logger

    logger = _logger.get_logger(__name__)
    return logger


class ADBClient(object):
    def __init__(self, identifier: Optional[str] = None):
        self._identifier = identifier
        self._busybox: Optional[optional.Optional[bool]] = None
        self._which: Optional[optional.Optional[bool]] = None

    @property
    def identifier(self):
        return self._identifier

    def has_busybox(self):
        if self._busybox is None:
            self._busybox = optional.Optional.of(self.exists(constants.BUSYBOX))
        return self._busybox.get()

    def hash_which(self):
        if self._which is None:
            self._which = optional.Optional.of(self.exists(constants.WHICH))
        return self._which

    def wait_for_device(self) -> bool:
        if self.is_connected():
            return True
        return adb_connection.wait_for_device(ip=self._identifier)

    def connect(self):
        return adb_connection.connect(ip=self._identifier)

    def disconnect(self):
        return adb_connection.disconnect(ip=self._identifier)

    @staticmethod
    def disconnect_all():
        return adb_connection.disconnect_all()

    def is_connected(self):
        return adb_connection.is_connected(ip=self._identifier)

    def reconnect(self):
        return adb_connection.reconnect(ip=self._identifier)

    def reconnect_device(self):
        return adb_connection.reconnect_device(ip=self._identifier)

    def reconnect_offline(self):
        return adb_connection.reconnect_offline(ip=self._identifier)

    def is_root(self):
        return adb_connection.is_root(ip=self._identifier)

    def root(self):
        self._connect_if_disconnected()
        if not self.is_root():
            adb_connection.root(ip=self._identifier)
            self.reconnect_device()
            self.wait_for_device()
            return adb_connection.is_root(ip=self._identifier)
        return True

    def unroot(self):
        self._connect_if_disconnected()
        if self.is_root():
            adb_connection.unroot(ip=self._identifier)
            self.reconnect_device()
            self.wait_for_device()
            return not adb_connection.is_root(ip=self._identifier)
        return True

    def shell(self, command: str, **kwargs):
        self._connect_if_disconnected()
        return adb_connection.shell(command=command, ip=self._identifier, **kwargs)

    def capture_output(self, command: str, **kwargs):
        self._connect_if_disconnected()
        return adb_connection.capture_output(
            command=command, ip=self._identifier, **kwargs
        )

    def execute(self, command: str, **kwargs):
        self._connect_if_disconnected()
        return adb_connection.execute(command=command, ip=self._identifier, **kwargs)

    def remount_filesystem(self, writeable: bool):
        self._connect_if_disconnected()
        return adb_connection.remount_as(ip=self._identifier, writeable=writeable)

    def remount(self):
        self._connect_if_disconnected()
        return adb_connection.remount(ip=self._identifier)

    def reboot(self, mode: Optional[str] = None, wait_device: bool = False):
        self._connect_if_disconnected()
        if adb_connection.reboot(ip=self._identifier, mode=mode):
            if wait_device:
                return self.wait_for_device()
            return True
        return False

    def bugreport(self, dst: Optional[str] = None):
        adb_connection.bugreport(dst, self._identifier)

    def get_id(self) -> Optional[str]:
        """
        Return the value of ANDROID_ID
        See: https://developer.android.com/reference/android/provider/Settings.Secure.html#ANDROID_ID

        :return:
        """
        code, output, _ = self.shell("settings get secure android_id")
        if code == ADBCommandResult.RESULT_OK:
            return output
        return None

    def getprop(
            self, key: Optional[str] = None
    ) -> Optional[Union[str, Dict[str, str]]]:
        if key:
            with self.shell(f"getprop {key}") as result:
                if result.is_ok():
                    return result.output()
                return None
        else:
            with self.shell("getprop") as result:
                if result.is_ok() and result.output():
                    return self._parse_properties(result.output())
                return None

    def setprop(self, key: str, value: str):
        return self.shell(f"setprop {key} {value}")

    def get_mac_address(self) -> Optional[str]:
        result = self.ifconfig()
        if result.code == ADBCommandResult.RESULT_OK:
            match = re.search("HWaddr[\\s]+([0-9:a-zA-Z]+)", result.output())
            if match and len(match.groups()) > 0:
                return match.groups()[0]
        return None

    def set_bluetooth_enabled(self, value: bool):
        log().verbose(f"Turning Bluetooth {'on' if value else 'off'}...")
        return self.shell(f"service call bluetooth_manager {6 if value else 8}")

    def get_global_settings(self) -> Optional[Dict[str, str]]:
        code, output, error = self.shell("settings list global")
        if code == ADBCommandResult.RESULT_OK and output:
            return dict(map(lambda x: x.split("=", 1), output.splitlines()))
        return None

    def get_global_settings_value(self, key: str):
        """Retrieve a value from the device global settings"""
        return self.shell(f"settings get global {key}")

    def set_global_settings_value(self, key: str, value: str):
        self.root()
        return self.shell(f"settings put global {key} {value}")

    def dumpsys_bluetooth(self):
        return self.dumpsys("bluetooth_manager")

    def dumpsys(self, which: str, **kwargs):
        """
        For a list of arguments see: https://developer.android.com/studio/command-line/dumpsys
        :param which:   which system service to dump
        :param kwargs:  additional arguments
        :return:
        """
        return self.shell(
            command="dumpsys", **adb_connection.extends_extra_arguments(which, **kwargs)
        )

    def dumpsys_meminfo(self, package: str):
        return self.shell(f"dumpsys meminfo {package}")

    def ifconfig(self):
        return self.shell("ifconfig")

    """ -----------------------------------------------------------------------"""
    """ File system methods """
    """ -----------------------------------------------------------------------"""

    def is_file(self, path: str) -> bool:
        return self._test_file(path, "f")

    def is_dir(self, path: str) -> bool:
        return self._test_file(path, "d")

    def is_symlink(self, path: str) -> bool:
        return self._test_file(path, "h")

    def exists(self, path: str) -> bool:
        return self._test_file(path, "e")

    def remove(self, dst: str):
        return self.shell(f"rm -fr {dst}")

    def cat(self, filename: str) -> str:
        code, output, error = self.shell(f"cat {filename}")
        if code != ADBCommandResult.RESULT_OK:
            raise IOError(f"File {filename} not found")
        return output

    def which(self, command: str) -> Optional[str]:
        self._connect_if_disconnected()
        if self.hash_which():
            return adb_connection.which(command=command, ip=self._identifier)
        elif self.has_busybox():
            result = self.busybox(f"which {command}")
            if result.is_ok():
                return result.output()
        else:
            return None

    def ls(self, dirname: str, **kwargs):
        args = adb_connection.expand_extra_arguments(**kwargs)
        args.append(dirname)
        return self.shell("ls", args=args)

    def busybox(self, command: str, **kwargs):
        return self.shell(f"busybox {command}", **kwargs)

    def push(self, src: Path, dst: str, **kwargs) -> bool:
        log().notice(f"--> Push `{src}` into `{dst}`...")
        self._connect_if_disconnected()
        return adb_connection.push(ip=self._identifier, src=str(src), dst=dst, **kwargs)

    def pull(self, src: str, dst: Path, **kwargs):
        log().notice(f"<-- Pull `{src}` into `{dst}`...")
        self._connect_if_disconnected()
        final_dst = dst
        if dst.is_dir():
            if not dst.exists():
                raise IOError(f"{dst} does not exist")
            final_dst = dst / os.path.basename(src)
        return adb_connection.pull(
            ip=self._identifier, src=src, dst=str(final_dst), **kwargs
        )

    """ -----------------------------------------------------------------------"""
    """ Key Events """
    """ -----------------------------------------------------------------------"""

    def getevent(self):
        result = self.shell("getevent -p")
        if result.is_ok():
            return re.findall(r"^add device [0-9]+: ([^\n]+)\n\s*name:\s*\"(.*)\"\n?", result.output(), re.MULTILINE)
        return None

    def getevent_named(self, name: str):
        events = self.getevent()
        if events and len(events) > 0:
            return next(map(lambda y: y[0], filter(lambda x: x[1] == name, events)), events[0])
        return '/dev/input/event0'

    def key_up(self, key: int, device: str = '/dev/input/event0'):
        return self.shell(
            f"""
        sendevent {device} 0004 0004 000c0042 &&
        sendevent {device} 0001 {key} 00000000 &&
        sendevent {device} 0000 0000 00000000
        """
        )

    def key_down(self, key: int, device: str = '/dev/input/event0'):
        return self.shell(
            f"""
        sendevent {device} 0004 0004 000c0042 &&
        sendevent {device} 0001 {key} 00000001 &&
        sendevent {device} 0000 0000 00000000
        """
        )

    def send_raw_key(self, key: int, device: str = '/dev/input/event0'):
        """
        Most values are from CplusPlusKeyCodes
        Seee https://supportcommunity.zebra.com/s/article/Find-out-key-event-via-ADB?language=en_US
        or
        http://androidxref.com/4.4_r1/xref/prebuilts/ndk/6/platforms/android-9/arch-arm/usr/include/linux/input.h
        for help about key codes
        """
        return self.shell(
            f"""
        sendevent {device} 0004 0004 000c0042 &&
        sendevent {device} 0001 {key} 00000001 &&
        sendevent {device} 0000 0000 00000000 &&
        sendevent {device} 0004 0004 000c0042 &&
        sendevent {device} 0001 {key} 00000000 &&
        sendevent {device} 0000 0000 00000000"""
        )

    def send_key(self, key: int) -> ADBCommandResult:
        assert type(key) is int
        log().verbose(f"Sending key: {key}")
        return self.shell(f"input keyevent {key}")

    def send_text(self, char: str) -> ADBCommandResult:
        assert type(char) is str
        log().verbose(f"Sending {char}...")
        return self.shell(f"input text {char}")

    """ ----------------------------------------------------------------------- """
    """ Private Methods """
    """ ----------------------------------------------------------------------- """

    def _test_file(self, path: str, mode: str) -> bool:
        result = self.shell("test -{} {} && echo 1 || echo 0".format(mode, path))
        if result.code == ADBCommandResult.RESULT_OK:
            return result.output().strip() == "1"
        return False

    def _connect_if_disconnected(self):
        if not self.is_connected():
            self.connect()

    def _parse_properties(self, string: str) -> Dict[str, str]:
        result = dict()
        lines = list(map(lambda x: x.split(":", 1), string.splitlines()))
        for line in lines:
            item = list(map(lambda y: y.strip()[1:-1], line))
            result[item[0]] = item[1]
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import enum

__all__ = ["KeyCodes", "CPlusPlusKeyCodes"]


class KeyCodes(enum.Enum):
    """
    These can be used when sending events with `adb shell input keyevent`\n
    See more https://developer.android.com/reference/android/view/KeyEvent
    """

    KEYCODE_0 = 7
    KEYCODE_11 = 227
    KEYCODE_12 = 228
    KEYCODE_1 = 8
    KEYCODE_2 = 9
    KEYCODE_3 = 10
    KEYCODE_3D_MODE = 206
    KEYCODE_4 = 11
    KEYCODE_5 = 12
    KEYCODE_6 = 13
    KEYCODE_7 = 14
    KEYCODE_8 = 15
    KEYCODE_9 = 16
    KEYCODE_A = 29
    KEYCODE_ALL_APPS = 284
    KEYCODE_ALT_LEFT = 57
    KEYCODE_ALT_RIGHT = 58
    KEYCODE_APOSTROPHE = 75
    KEYCODE_APP_SWITCH = 187
    KEYCODE_ASSIST = 219
    KEYCODE_AT = 77
    KEYCODE_AVR_INPUT = 182
    KEYCODE_AVR_POWER = 181
    KEYCODE_B = 30
    KEYCODE_BACK = 4
    KEYCODE_BACKSLASH = 73
    KEYCODE_BOOKMARK = 174
    KEYCODE_BREAK = 121
    KEYCODE_BRIGHTNESS_DOWN = 220
    KEYCODE_BRIGHTNESS_UP = 221
    KEYCODE_BUTTON_10 = 197
    KEYCODE_BUTTON_11 = 198
    KEYCODE_BUTTON_12 = 199
    KEYCODE_BUTTON_13 = 200
    KEYCODE_BUTTON_14 = 201
    KEYCODE_BUTTON_15 = 202
    KEYCODE_BUTTON_16 = 203
    KEYCODE_BUTTON_1 = 188
    KEYCODE_BUTTON_2 = 189
    KEYCODE_BUTTON_3 = 190
    KEYCODE_BUTTON_4 = 191
    KEYCODE_BUTTON_5 = 192
    KEYCODE_BUTTON_6 = 193
    KEYCODE_BUTTON_7 = 194
    KEYCODE_BUTTON_8 = 195
    KEYCODE_BUTTON_9 = 196
    KEYCODE_BUTTON_A = 96
    KEYCODE_BUTTON_B = 97
    KEYCODE_BUTTON_C = 98
    KEYCODE_BUTTON_L1 = 102
    KEYCODE_BUTTON_L2 = 104
    KEYCODE_BUTTON_MODE = 110
    KEYCODE_BUTTON_R1 = 103
    KEYCODE_BUTTON_R2 = 105
    KEYCODE_BUTTON_SELECT = 109
    KEYCODE_BUTTON_START = 108
    KEYCODE_BUTTON_THUMBL = 106
    KEYCODE_BUTTON_THUMBR = 107
    KEYCODE_BUTTON_X = 99
    KEYCODE_BUTTON_Y = 100
    KEYCODE_BUTTON_Z = 101
    KEYCODE_C = 31
    KEYCODE_CALCULATOR = 210
    KEYCODE_CALENDAR = 208
    KEYCODE_CALL = 5
    KEYCODE_CAMERA = 27
    KEYCODE_CAPS_LOCK = 115
    KEYCODE_CAPTIONS = 175
    KEYCODE_CHANNEL_DOWN = 167
    KEYCODE_CHANNEL_UP = 166
    KEYCODE_CLEAR = 28
    KEYCODE_COMMA = 55
    KEYCODE_CONTACTS = 207
    KEYCODE_COPY = 278
    KEYCODE_CTRL_LEFT = 113
    KEYCODE_CTRL_RIGHT = 114
    KEYCODE_CUT = 277
    KEYCODE_D = 32
    KEYCODE_DEL = 67
    KEYCODE_DPAD_CENTER = 23
    KEYCODE_DPAD_DOWN = 20
    KEYCODE_DPAD_DOWN_LEFT = 269
    KEYCODE_DPAD_DOWN_RIGHT = 271
    KEYCODE_DPAD_LEFT = 21
    KEYCODE_DPAD_RIGHT = 22
    KEYCODE_DPAD_UP = 19
    KEYCODE_DPAD_UP_LEFT = 268
    KEYCODE_DPAD_UP_RIGHT = 270
    KEYCODE_DVR = 173
    KEYCODE_E = 33
    KEYCODE_EISU = 212
    KEYCODE_ENDCALL = 6
    KEYCODE_ENTER = 66
    KEYCODE_ENVELOPE = 65
    KEYCODE_EQUALS = 70
    KEYCODE_ESCAPE = 111
    KEYCODE_EXPLORER = 64
    KEYCODE_F10 = 140
    KEYCODE_F11 = 141
    KEYCODE_F12 = 142
    KEYCODE_F1 = 131
    KEYCODE_F2 = 132
    KEYCODE_F3 = 133
    KEYCODE_F4 = 134
    KEYCODE_F5 = 135
    KEYCODE_F6 = 136
    KEYCODE_F7 = 137
    KEYCODE_F8 = 138
    KEYCODE_F9 = 139
    KEYCODE_F = 34
    KEYCODE_FOCUS = 80
    KEYCODE_FORWARD = 125
    KEYCODE_FORWARD_DEL = 112
    KEYCODE_FUNCTION = 119
    KEYCODE_G = 35
    KEYCODE_GRAVE = 68
    KEYCODE_GUIDE = 172
    KEYCODE_H = 36
    KEYCODE_HEADSETHOOK = 79
    KEYCODE_HELP = 259
    KEYCODE_HENKAN = 214
    KEYCODE_HOME = 3
    KEYCODE_I = 37
    KEYCODE_INFO = 165
    KEYCODE_INSERT = 124
    KEYCODE_J = 38
    KEYCODE_K = 39
    KEYCODE_KANA = 218
    KEYCODE_KATAKANA_HIRAGANA = 215
    KEYCODE_L = 40
    KEYCODE_LANGUAGE_SWITCH = 204
    KEYCODE_LAST_CHANNEL = 229
    KEYCODE_LEFT_BRACKET = 71
    KEYCODE_M = 41
    KEYCODE_MANNER_MODE = 205
    KEYCODE_MEDIA_AUDIO_TRACK = 222
    KEYCODE_MEDIA_CLOSE = 128
    KEYCODE_MEDIA_EJECT = 129
    KEYCODE_MEDIA_FAST_FORWARD = 90
    KEYCODE_MEDIA_NEXT = 87
    KEYCODE_MEDIA_PAUSE = 127
    KEYCODE_MEDIA_PLAY = 126
    KEYCODE_MEDIA_PLAY_PAUSE = 85
    KEYCODE_MEDIA_PREVIOUS = 88
    KEYCODE_MEDIA_RECORD = 130
    KEYCODE_MEDIA_REWIND = 89
    KEYCODE_MEDIA_SKIP_BACKWARD = 273
    KEYCODE_MEDIA_SKIP_FORWARD = 272
    KEYCODE_MEDIA_STEP_BACKWARD = 275
    KEYCODE_MEDIA_STEP_FORWARD = 274
    KEYCODE_MEDIA_STOP = 86
    KEYCODE_MEDIA_TOP_MENU = 226
    KEYCODE_MENU = 82
    KEYCODE_META_LEFT = 117
    KEYCODE_META_RIGHT = 118
    KEYCODE_MINUS = 69
    KEYCODE_MOVE_END = 123
    KEYCODE_MOVE_HOME = 122
    KEYCODE_MUHENKAN = 213
    KEYCODE_MUSIC = 209
    KEYCODE_MUTE = 91
    KEYCODE_N = 42
    KEYCODE_NAVIGATE_IN = 262
    KEYCODE_NAVIGATE_NEXT = 261
    KEYCODE_NAVIGATE_OUT = 263
    KEYCODE_NAVIGATE_PREVIOUS = 260
    KEYCODE_NOTIFICATION = 83
    KEYCODE_NUM = 78
    KEYCODE_NUM_LOCK = 143
    KEYCODE_NUMPAD_0 = 144
    KEYCODE_NUMPAD_1 = 145
    KEYCODE_NUMPAD_2 = 146
    KEYCODE_NUMPAD_3 = 147
    KEYCODE_NUMPAD_4 = 148
    KEYCODE_NUMPAD_5 = 149
    KEYCODE_NUMPAD_6 = 150
    KEYCODE_NUMPAD_7 = 151
    KEYCODE_NUMPAD_8 = 152
    KEYCODE_NUMPAD_9 = 153
    KEYCODE_NUMPAD_ADD = 157
    KEYCODE_NUMPAD_COMMA = 159
    KEYCODE_NUMPAD_DIVIDE = 154
    KEYCODE_NUMPAD_DOT = 158
    KEYCODE_NUMPAD_ENTER = 160
    KEYCODE_NUMPAD_EQUALS = 161
    KEYCODE_NUMPAD_LEFT_PAREN = 162
    KEYCODE_NUMPAD_MULTIPLY = 155
    KEYCODE_NUMPAD_RIGHT_PAREN = 163
    KEYCODE_NUMPAD_SUBTRACT = 156
    KEYCODE_O = 43
    KEYCODE_P = 44
    KEYCODE_PAGE_DOWN = 93
    KEYCODE_PAGE_UP = 92
    KEYCODE_PAIRING = 225
    KEYCODE_PASTE = 279
    KEYCODE_PERIOD = 56
    KEYCODE_PICTSYMBOLS = 94
    KEYCODE_PLUS = 81
    KEYCODE_POUND = 18
    KEYCODE_POWER = 26
    KEYCODE_PROFILE_SWITCH = 288
    KEYCODE_PROG_BLUE = 186
    KEYCODE_PROG_GREEN = 184
    KEYCODE_PROG_RED = 183
    KEYCODE_PROG_YELLOW = 185
    KEYCODE_Q = 45
    KEYCODE_R = 46
    KEYCODE_REFRESH = 285
    KEYCODE_RIGHT_BRACKET = 72
    KEYCODE_RO = 217
    KEYCODE_S = 47
    KEYCODE_SCROLL_LOCK = 116
    KEYCODE_SEARCH = 84
    KEYCODE_SEMICOLON = 74
    KEYCODE_SETTINGS = 176
    KEYCODE_SHIFT_LEFT = 59
    KEYCODE_SHIFT_RIGHT = 60
    KEYCODE_SLASH = 76
    KEYCODE_SLEEP = 223
    KEYCODE_SOFT_LEFT = 1
    KEYCODE_SOFT_RIGHT = 2
    KEYCODE_SOFT_SLEEP = 276
    KEYCODE_SPACE = 62
    KEYCODE_STAR = 17
    KEYCODE_STB_INPUT = 180
    KEYCODE_STB_POWER = 179
    KEYCODE_STEM_1 = 265
    KEYCODE_STEM_2 = 266
    KEYCODE_STEM_3 = 267
    KEYCODE_STEM_PRIMARY = 264
    KEYCODE_SWITCH_CHARSET = 95
    KEYCODE_SYM = 63
    KEYCODE_SYSRQ = 120
    KEYCODE_SYSTEM_NAVIGATION_DOWN = 281
    KEYCODE_SYSTEM_NAVIGATION_LEFT = 282
    KEYCODE_SYSTEM_NAVIGATION_RIGHT = 283
    KEYCODE_SYSTEM_NAVIGATION_UP = 280
    KEYCODE_T = 48
    KEYCODE_TAB = 61
    KEYCODE_THUMBS_DOWN = 287
    KEYCODE_THUMBS_UP = 286
    KEYCODE_TV = 170
    KEYCODE_TV_ANTENNA_CABLE = 242
    KEYCODE_TV_AUDIO_DESCRIPTION = 252
    KEYCODE_TV_AUDIO_DESCRIPTION_MIX_DOWN = 254
    KEYCODE_TV_AUDIO_DESCRIPTION_MIX_UP = 253
    KEYCODE_TV_CONTENTS_MENU = 256
    KEYCODE_TV_DATA_SERVICE = 230
    KEYCODE_TV_INPUT = 178
    KEYCODE_TV_INPUT_COMPONENT_1 = 249
    KEYCODE_TV_INPUT_COMPONENT_2 = 250
    KEYCODE_TV_INPUT_COMPOSITE_1 = 247
    KEYCODE_TV_INPUT_COMPOSITE_2 = 248
    KEYCODE_TV_INPUT_HDMI_1 = 243
    KEYCODE_TV_INPUT_HDMI_2 = 244
    KEYCODE_TV_INPUT_HDMI_3 = 245
    KEYCODE_TV_INPUT_HDMI_4 = 246
    KEYCODE_TV_INPUT_VGA_1 = 251
    KEYCODE_TV_MEDIA_CONTEXT_MENU = 257
    KEYCODE_TV_NETWORK = 241
    KEYCODE_TV_NUMBER_ENTRY = 234
    KEYCODE_TV_POWER = 177
    KEYCODE_TV_RADIO_SERVICE = 232
    KEYCODE_TV_SATELLITE = 237
    KEYCODE_TV_SATELLITE_BS = 238
    KEYCODE_TV_SATELLITE_CS = 239
    KEYCODE_TV_SATELLITE_SERVICE = 240
    KEYCODE_TV_TELETEXT = 233
    KEYCODE_TV_TERRESTRIAL_ANALOG = 235
    KEYCODE_TV_TERRESTRIAL_DIGITAL = 236
    KEYCODE_TV_TIMER_PROGRAMMING = 258
    KEYCODE_TV_ZOOM_MODE = 255
    KEYCODE_U = 49
    KEYCODE_UNKNOWN = 0
    KEYCODE_V = 50
    KEYCODE_VOICE_ASSIST = 231
    KEYCODE_VOLUME_DOWN = 25
    KEYCODE_VOLUME_MUTE = 164
    KEYCODE_VOLUME_UP = 24
    KEYCODE_W = 51
    KEYCODE_WAKEUP = 224
    KEYCODE_WINDOW = 171
    KEYCODE_X = 52
    KEYCODE_Y = 53
    KEYCODE_YEN = 216
    KEYCODE_Z = 54
    KEYCODE_ZENKAKU_HANKAKU = 211
    KEYCODE_ZOOM_IN = 168
    KEYCODE_ZOOM_OUT = 169


class CPlusPlusKeyCodes(enum.Enum):
    """This can be used when sending events with `adb shell sendevent /dev/sendevent /dev/input/event3`"""

    KEY_0 = 11
    KEY_0_5 = 234
    KEY_1 = 2
    KEY_102ND = 86
    KEY_2 = 3
    KEY_2_5 = 235
    KEY_3 = 4
    KEY_4 = 5
    KEY_5 = 6
    KEY_6 = 7
    KEY_7 = 8
    KEY_8 = 9
    KEY_9 = 10
    KEY_A = 30
    KEY_AB = 0x196
    KEY_AGAIN = 129
    KEY_ALTERASE = 222
    KEY_ANGLE = 0x173
    KEY_APOSTROPHE = 40
    KEY_ARCHIVE = 0x169
    KEY_AUDIO = 0x188
    KEY_AUX = 0x186
    KEY_B = 48
    KEY_BACK = 158
    KEY_BACKSLASH = 43
    KEY_BACKSPACE = 14
    KEY_BASSBOOST = 209
    KEY_BATTERY = 236
    KEY_BLUE = 0x191
    KEY_BOOKMARKS = 156
    KEY_BREAK = 0x19B
    KEY_BRIGHTNESSDOWN = 224
    KEY_BRIGHTNESSUP = 225
    KEY_BRL_DOT1 = 0x1F1
    KEY_BRL_DOT2 = 0x1F2
    KEY_BRL_DOT3 = 0x1F3
    KEY_BRL_DOT4 = 0x1F4
    KEY_BRL_DOT5 = 0x1F5
    KEY_BRL_DOT6 = 0x1F6
    KEY_BRL_DOT7 = 0x1F7
    KEY_BRL_DOT8 = 0x1F8
    KEY_C = 46
    KEY_CALC = 140
    KEY_CALENDAR = 0x18D
    KEY_CAMERA = 212
    KEY_CANCEL = 223
    KEY_CAPSLOCK = 58
    KEY_CD = 0x17F
    KEY_CENTER = 232
    KEY_CHANNEL = 0x16B
    KEY_CHANNELDOWN = 0x193
    KEY_CHANNELUP = 0x192
    KEY_CHAT = 216
    KEY_CLEAR = 0x163
    KEY_CLOSE = 206
    KEY_CLOSECD = 160
    KEY_COFFEE = 152
    KEY_COMMA = 51
    KEY_COMPOSE = 127
    KEY_COMPUTER = 157
    KEY_CONFIG = 171
    KEY_CONNECT = 218
    KEY_COPY = 133
    KEY_CUT = 137
    KEY_CYCLEWINDOWS = 154
    KEY_D = 32
    KEY_DEL_EOL = 0x1C0
    KEY_DEL_EOS = 0x1C1
    KEY_DEL_LINE = 0x1C3
    KEY_DELETE = 111
    KEY_DELETEFILE = 146
    KEY_DIGITS = 0x19D
    KEY_DIRECTION = 153
    KEY_DIRECTORY = 0x18A
    KEY_DOCUMENTS = 235
    KEY_DOT = 52
    KEY_DOWN = 108
    KEY_DVD = 0x185
    KEY_E = 18
    KEY_EDIT = 176
    KEY_EJECTCD = 161
    KEY_EJECTCLOSECD = 162
    KEY_EMAIL = 215
    KEY_END = 107
    KEY_ENTER = 28
    KEY_EPG = 0x16D
    KEY_EQUAL = 13
    KEY_ESC = 1
    KEY_EXIT = 174
    KEY_F = 33
    KEY_F1 = 59
    KEY_F10 = 68
    KEY_F11 = 87
    KEY_F12 = 88
    KEY_F13 = 183
    KEY_F14 = 184
    KEY_F15 = 185
    KEY_F16 = 186
    KEY_F17 = 187
    KEY_F18 = 188
    KEY_F19 = 189
    KEY_F2 = 60
    KEY_F20 = 190
    KEY_F21 = 191
    KEY_F22 = 192
    KEY_F23 = 193
    KEY_F24 = 194
    KEY_F3 = 61
    KEY_F4 = 62
    KEY_F5 = 63
    KEY_F6 = 64
    KEY_F7 = 65
    KEY_F8 = 66
    KEY_F9 = 67
    KEY_FASTFORWARD = 208
    KEY_FAVORITES = 0x16C
    KEY_FILE = 144
    KEY_FINANCE = 219
    KEY_FIND = 136
    KEY_FIRST = 0x194
    KEY_FN = 0x1D0
    KEY_FN_1 = 0x1DE
    KEY_FN_2 = 0x1DF
    KEY_FN_B = 0x1E4
    KEY_FN_D = 0x1E0
    KEY_FN_E = 0x1E1
    KEY_FN_ESC = 0x1D1
    KEY_FN_F = 0x1E2
    KEY_FN_F1 = 0x1D2
    KEY_FN_F10 = 0x1DB
    KEY_FN_F11 = 0x1DC
    KEY_FN_F12 = 0x1DD
    KEY_FN_F2 = 0x1D3
    KEY_FN_F3 = 0x1D4
    KEY_FN_F4 = 0x1D5
    KEY_FN_F5 = 0x1D6
    KEY_FN_F6 = 0x1D7
    KEY_FN_F7 = 0x1D8
    KEY_FN_F8 = 0x1D9
    KEY_FN_F9 = 0x1DA
    KEY_FN_S = 0x1E3
    KEY_FORWARD = 159
    KEY_FORWARDMAIL = 233
    KEY_FRONT = 132
    KEY_G = 34
    KEY_GOTO = 0x162
    KEY_GRAVE = 41
    KEY_GREEN = 0x18F
    KEY_H = 35
    KEY_HANGEUL = 122
    KEY_HANJA = 123
    KEY_HEADSETHOOK = 233
    KEY_HELP = 138
    KEY_HENKAN = 92
    KEY_HIRAGANA = 91
    KEY_HOME = 102
    KEY_HOMEPAGE = 172
    KEY_HP = 211
    KEY_I = 23
    KEY_INFO = 0x166
    KEY_INS_LINE = 0x1C2
    KEY_INSERT = 110
    KEY_ISO = 170
    KEY_J = 36
    KEY_K = 37
    KEY_KATAKANA = 90
    KEY_KATAKANAHIRAGANA = 93
    KEY_KBDILLUMDOWN = 238
    KEY_KBDILLUMTOGGLE = 237
    KEY_KBDILLUMUP = 239
    KEY_KEYBOARD = 0x176
    KEY_KP0 = 82
    KEY_KP1 = 79
    KEY_KP2 = 80
    KEY_KP3 = 81
    KEY_KP4 = 75
    KEY_KP5 = 76
    KEY_KP6 = 77
    KEY_KP7 = 71
    KEY_KP8 = 72
    KEY_KP9 = 73
    KEY_KPASTERISK = 55
    KEY_KPCOMMA = 121
    KEY_KPDOT = 83
    KEY_KPENTER = 96
    KEY_KPEQUAL = 117
    KEY_KPJPCOMMA = 95
    KEY_KPLEFTPAREN = 179
    KEY_KPMINUS = 74
    KEY_KPPLUS = 78
    KEY_KPPLUSMINUS = 118
    KEY_KPRIGHTPAREN = 180
    KEY_KPSLASH = 98
    KEY_L = 38
    KEY_LANGUAGE = 0x170
    KEY_LAST = 0x195
    KEY_LEFT = 105
    KEY_LEFTALT = 56
    KEY_LEFTBRACE = 26
    KEY_LEFTCTRL = 29
    KEY_LEFTMETA = 125
    KEY_LEFTSHIFT = 42
    KEY_LINEFEED = 101
    KEY_LIST = 0x18B
    KEY_M = 50
    KEY_MACRO = 112
    KEY_MAIL = 155
    KEY_MEDIA = 226
    KEY_MEMO = 0x18C
    KEY_MENU = 139
    KEY_MHP = 0x16F
    KEY_MINUS = 12
    KEY_MODE = 0x175
    KEY_MOVE = 175
    KEY_MP3 = 0x187
    KEY_MSDOS = 151
    KEY_MUHENKAN = 94
    KEY_MUTE = 113
    KEY_N = 49
    KEY_NEW = 181
    KEY_NEXT = 0x197
    KEY_NEXTSONG = 163
    KEY_NUMLOCK = 69
    KEY_O = 24
    KEY_OK = 0x160
    KEY_OPEN = 134
    KEY_OPTION = 0x165
    KEY_P = 25
    KEY_PAGEDOWN = 109
    KEY_PAGEUP = 104
    KEY_PASTE = 135
    KEY_PAUSE = 119
    KEY_PAUSECD = 201
    KEY_PC = 0x178
    KEY_PHONE = 169
    KEY_PLAY = 207
    KEY_PLAYCD = 200
    KEY_PLAYER = 0x183
    KEY_PLAYPAUSE = 164
    KEY_POWER = 116
    KEY_POWER2 = 0x164
    KEY_PREVIOUS = 0x19C
    KEY_PREVIOUSSONG = 165
    KEY_PRINT = 210
    KEY_PROG1 = 148
    KEY_PROG2 = 149
    KEY_PROG3 = 202
    KEY_PROG4 = 203
    KEY_PROGRAM = 0x16A
    KEY_PROPS = 130
    KEY_PVR = 0x16E
    KEY_Q = 16
    KEY_QUESTION = 214
    KEY_R = 19
    KEY_RECORD = 167
    KEY_RED = 0x18E
    KEY_REDO = 182
    KEY_REFRESH = 173
    KEY_REPLY = 232
    KEY_RESERVED = 0
    KEY_RESTART = 0x198
    KEY_REWIND = 168
    KEY_RIGHT = 106
    KEY_RIGHTALT = 100
    KEY_RIGHTBRACE = 27
    KEY_RIGHTCTRL = 97
    KEY_RIGHTMETA = 126
    KEY_RIGHTSHIFT = 54
    KEY_RO = 89
    KEY_S = 31
    KEY_SAT = 0x17D
    KEY_SAT2 = 0x17E
    KEY_SAVE = 234
    KEY_SCREEN = 0x177
    KEY_SCROLLDOWN = 178
    KEY_SCROLLLOCK = 70
    KEY_SCROLLUP = 177
    KEY_SEARCH = 217
    KEY_SELECT = 0x161
    KEY_SEMICOLON = 39
    KEY_SEND = 231
    KEY_SENDFILE = 145
    KEY_SETUP = 141
    KEY_SHARP = 228
    KEY_SHOP = 221
    KEY_SHUFFLE = 0x19A
    KEY_SLASH = 53
    KEY_SLEEP = 142
    KEY_SLOW = 0x199
    KEY_SOFT1 = 229
    KEY_SOFT2 = 230
    KEY_SOUND = 213
    KEY_SPACE = 57
    KEY_SPORT = 220
    KEY_STAR = 227
    KEY_STOP = 128
    KEY_STOPCD = 166
    KEY_SUBTITLE = 0x172
    KEY_SUSPEND = 205
    KEY_SWITCHVIDEOMODE = 236
    KEY_SYSRQ = 99
    KEY_T = 20
    KEY_TAB = 15
    KEY_TAPE = 0x180
    KEY_TEEN = 0x19E
    KEY_TEXT = 0x184
    KEY_TIME = 0x167
    KEY_TITLE = 0x171
    KEY_TUNER = 0x182
    KEY_TV = 0x179
    KEY_TV2 = 0x17A
    KEY_TWEN = 0x19F
    KEY_U = 22
    KEY_UNDO = 131
    KEY_UNKNOWN = 240
    KEY_UP = 103
    KEY_V = 47
    KEY_VCR = 0x17B
    KEY_VCR2 = 0x17C
    KEY_VENDOR = 0x168
    KEY_VIDEO = 0x189
    KEY_VOLUMEDOWN = 114
    KEY_VOLUMEUP = 115
    KEY_W = 17
    KEY_WAKEUP = 143
    KEY_WWW = 150
    KEY_X = 45
    KEY_XFER = 147
    KEY_Y = 21
    KEY_YELLOW = 0x190
    KEY_YEN = 124
    KEY_Z = 44
    KEY_ZENKAKUHANKAKU = 85
    KEY_ZOOM = 0x174
    KEY_RADIO = 386



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import re
import tempfile
from pathlib import Path
from typing import Dict, List, Optional

from xml.etree import ElementTree

from . import ADBClient
from . import Package
from .adb_connection import ADBCommandResult, extends_extra_arguments

__all__ = ["PackageManager"]

# noinspection RegExpRedundantEscape
package_pattern = re.compile(
    r"package:(?P<apk>.*\.apk)?=?(?P<package>[a-zA-Z0-9\.]+)\s*(installer=(?P<installer>[\w\.]+))?\s*(uid:(?P<uuid>[\w]+))?"
)


def log():
    from . import _logger
    return _logger.get_logger(__name__)


class PackageManager(object):
    def __init__(self, client: ADBClient):
        self._client = client
        self._has_cmd = False
        self._has_cmd_checked = False

    @property
    def client(self):
        return self._client

    def dump(self, package: str):
        if not self.is_installed(package):
            return ADBCommandResult.from_error()
        return self.client.shell(f"pm dump {package}")

    @property
    def has_cmd(self):
        if not self._has_cmd_checked:
            self._has_cmd = self.client.which("cmd") is not None
            self._has_cmd_checked = True
        return self._has_cmd

    """ -----------------------------------------------------------------------"""
    """ Permissions """
    """ -----------------------------------------------------------------------"""

    def grant_permission(self, package: str, permission: str):
        log().verbose(f"Granting '{permission}' to '{package}'...")
        return self.client.shell(f"pm grant {package} {permission}")

    def revoke_permission(self, package: str, permission: str):
        log().verbose(f"Revoking '{permission}' from '{package}'...")
        return self.client.shell(f"pm revoke {package} {permission}")

    def has_runtime_permission(self, package: str, permission: str) -> bool:
        """Test if the given package has the runtime permission granted"""
        return any(
            x[0] == permission and x[1]
            for x in self.get_runtime_permissions(package).items()
        )

    def get_runtime_permissions(self, package: str) -> Dict[str, bool]:
        """
        Returns the package runtime permissions
        :param package:
        :return:
        """
        log().verbose(f"Fetching runtime permissions for `{package}`...")
        code, output, error = self.dump(package)
        if code == ADBCommandResult.RESULT_OK and output:
            return _parse_dump_permissions("runtime permissions:", output)
        return dict()

    def get_requested_permissions(self, package: str) -> List[str]:
        """
        Returns the package runtime permissions
        :param package:
        :return:
        """
        log().verbose(f"Fetching requested permissions for `{package}`...")
        code, output, error = self.dump(package)
        if code == ADBCommandResult.RESULT_OK and output:
            return list(
                _parse_dump_permissions("requested permissions:", output).keys()
            )
        return list()

    def get_install_permissions(self, package: str) -> Dict[str, bool]:
        """
        Returns the package runtime permissions
        :param package:
        :return:
        """
        log().verbose(f"Fetching install permissions for `{package}`...")
        code, output, error = self.dump(package)
        if code == ADBCommandResult.RESULT_OK and output:
            return _parse_dump_permissions("install permissions:", output)
        return dict()

    """ -----------------------------------------------------------------------"""
    """ Install/Uninstall/Clear """
    """ -----------------------------------------------------------------------"""

    def install(self, apk: Path, **kwargs) -> ADBCommandResult:
        log().verbose(f"install {apk}")
        return self._install_package(apk, **kwargs)

    def uninstall(self, package: str) -> ADBCommandResult:
        """
        uninstall [-k] PACKAGE
            remove this app package from the device
            '-k': keep the data and cache directories
        """
        log().verbose(f"uninstall {package}")
        if self.is_installed(package):
            if self.is_system(package):
                return self._uninstall_package(package, args=("--user 0",))
            else:
                return self._uninstall_package(package)
        else:
            log().warning(f"{package} is not installed")
            return ADBCommandResult.from_error()

    def remove_entry(self, package: str) -> bool:
        """
        Remove the package entry from packages.xml and packages.list inside /data/system
        :param package:
        :return:
        """
        return _remove_package_from_packages(self.client, package)

    def clear(self, package: str) -> ADBCommandResult:
        log().verbose(f"clear {package}")
        if self.is_installed(package):
            if self.is_system(package):
                return self._clear_package(package, args=("--user 0",))
            else:
                return self._clear_package(package)
        else:
            log().warning(f"{package} is not installed")
            return ADBCommandResult.from_error()

    def forcestop(self, package: str):
        return self.client.shell(f"am force-stop {package}")

    def install_multi_package(self, *apks, **kwargs):
        """
        Not available on all adb clients

        install-multi-package [-lrtsdpg] [--instant] PACKAGE...
            push one or more packages to the device and install them atomically
            -r: replace existing application
            -t: allow test packages
            -d: allow version code downgrade (debuggable packages only)
            -p: partial application install (install-multiple only)
            -g: grant all runtime permissions
            --abi ABI: override platform's default ABI
            --instant: cause the app to be installed as an ephemeral install app
            --no-streaming: always push APK to device and invoke Package Manager as separate steps
            --streaming: force streaming APK directly into Package Manager
            --fastdeploy: use fast deploy
            --no-fastdeploy: prevent use of fast deploy
            --force-agent: force update of deployment agent when using fast deploy
            --date-check-agent: update deployment agent when local version is newer and using fast deploy
            --version-check-agent: update deployment agent when local version has different version code and using fast deploy
            --local-agent: locate agent files from local source build (instead of SDK location)
            (See also `adb shell pm help` for more options.)
        :param apks:
        :param kwargs:
        :return:
        """
        return self.client.capture_output(
            command="install-multi-package",
            **extends_extra_arguments(*list(map(lambda x: str(x), apks)), **kwargs),
        )

    """
    """

    def list(self, package: Optional[str] = None, **kwargs):
        if self.has_cmd:
            return _cmd_list(self.client, package, **kwargs)
        return _pm_list(self.client, package, **kwargs)

    def is_installed(self, package: str) -> bool:
        return any(x.name == package for x in self.list(package))

    def find(self, package: str) -> Optional[Package]:
        result = self.list(package, args=("-f", "-U", "-i"))
        log().spam(f"result: {result}")
        return next((x for x in result if x.name == package), None)

    def is_system(self, package: str):
        result = self.find(package)
        if result is not None:
            return result.is_system()
        return False

    """ -----------------------------------------------------------------------"""
    """ Private Methods """
    """ -----------------------------------------------------------------------"""

    def _uninstall_package(self, package: str, **kwargs) -> ADBCommandResult:
        return self.client.shell(
            "pm uninstall",
            **extends_extra_arguments(str(package), **kwargs),
        )

    def _clear_package(self, package: str, **kwargs) -> ADBCommandResult:
        return self.client.shell(
            "pm clear", **extends_extra_arguments(str(package), **kwargs)
        )

    def _install_package(self, apk: Path, **kwargs) -> ADBCommandResult:
        return self.client.capture_output(
            command="install",
            **extends_extra_arguments(str(apk), **kwargs),
        )


def _parse_dump_permissions(headline: str, input_str: str) -> Dict[str, bool]:
    found = False
    leading_spaces = 0
    permissions = dict()
    for line in input_str.splitlines():
        if line.strip() == headline:
            leading_spaces = len(line) - len(line.lstrip())
            found = True
            continue

        if found:
            if len(line) - len(line.lstrip()) == leading_spaces:
                found = False
                continue

            line = line.strip()
            if len(line) == 0:
                found = False
                continue

            splitted_line = line.split(":")
            if len(splitted_line) == 2:
                name = splitted_line[0]
                granted = splitted_line[1].split("=")[1].strip()
                permissions[name] = granted == "true"
            else:
                permissions[splitted_line[0]] = False
    return permissions


def _remove_package_from_packages(client: ADBClient, package) -> bool:
    if not _remove_entry_from_packages_list(client, package):
        return False
    return _remove_entry_from_packages_xml(client, package)


def _remove_entry_from_packages_list(client: ADBClient, package) -> bool:
    log().debug("Removing entry from '/data/system/packages.list'...")
    content = client.cat("/data/system/packages.list")
    splitted_content = list(filter(lambda x: not x.startswith(package), content.splitlines()))
    local_file = Path(tempfile.gettempdir()) / "packages.list"
    with open(local_file, "w") as fp:
        for line in splitted_content:
            fp.write(line)
        fp.close()

    if not client.push(Path(local_file), "/data/system/"):
        log().warning("Failed to push '%s' to '/data/system/'" % local_file)
        return False
    return True


def _remove_entry_from_packages_xml(client: ADBClient, package) -> bool:
    log().debug("Removing entry from '/data/system/packages.xml'...")
    content = client.cat("/data/system/packages.xml")

    root = ElementTree.ElementTree(ElementTree.fromstring(content))
    for elem in root.findall(".//package[@name='%s']" % package):
        log().verbose(f"Removing {elem} from node")
        root.getroot().remove(elem)

    local_file = Path(tempfile.gettempdir()) / "packages.xml"

    with open(local_file, "wb") as fp:
        root.write(fp, encoding="utf-8")

    if not client.push(Path(local_file), "/data/system/"):
        log().warning(f"Failed to push '{local_file}' to '/data/system/'")
        return False
    return True


def _pm_list(client: ADBClient, package: Optional[str] = None, **kwargs):
    """
    Uses legacy "pm list packages"
    pm list packages: prints all packages, optionally only
    those whose package name contains the text in FILTER.  Options:
        -f: see their associated file.
        -d: filter to only show disbled packages.
        -e: filter to only show enabled packages.
        -s: filter to only show system packages.
        -3: filter to only show third party packages.
        -i: see the installer for the packages.
        -u: also include uninstalled packages.
    :param client: current ADBClient
    :param package:
    :param kwargs:
    :return:
    """
    result = client.shell(
        "pm list packages",
        **extends_extra_arguments(f"{package if package else ''}", **kwargs),
    )
    if result.is_ok() and result.output():
        return _parse_packages(result.output())
    return list()


def _cmd_list(client: ADBClient, package: Optional[str] = None, **kwargs):
    """
    list packages [-f] [-d] [-e] [-s] [-3] [-i] [-l] [-u] [-U] [--uid UID] [--user USER_ID] [FILTER]
    Prints all packages; optionally only those whose name contains
    the text in FILTER.
    Options:
    -f: see their associated file
    -d: filter to only show disabled packages
    -e: filter to only show enabled packages
    -s: filter to only show system packages
    -3: filter to only show third party packages
    -i: see the installer for the packages
    -l: ignored (used for compatibility with older releases)
    -U: also show the package UID
    -u: also include uninstalled packages
    --uid UID: filter to only show packages with the given UID
    --user USER_ID: only list packages belonging to the given user

    :param client: current ADBClient
    :param package: optional package to look for
    :param kwargs:
    :return:
    """
    result = client.shell(
        "cmd package list packages",
        **extends_extra_arguments(f"{package if package else ''}", **kwargs),
    )
    if result.is_ok() and result.output():
        return _parse_packages(result.output())
    return list()


def _parse_packages(output: str) -> List[Package]:
    packages = list()
    for line in output.splitlines():
        match = re.match(package_pattern, line)
        if match:
            packages.append(Package(match.groupdict()))
    return packages



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["PropertyParser"]

from typing import KeysView, Optional, ItemsView


class PropertyParser:
    def __init__(self):
        self.propDict = dict()

    def read_string(self, string: str) -> PropertyParser:
        for propLine in string.splitlines():
            prop_def = propLine.strip()
            if len(prop_def) == 0:
                continue
            if prop_def[0] in ("!", "#"):
                continue
            punctuation = [prop_def.find(c) for c in ":= "] + [len(prop_def)]
            found = min([pos for pos in punctuation if pos != -1])
            name = prop_def[:found].rstrip()
            value = prop_def[found:].lstrip(":= ").rstrip()
            self.propDict[name] = value
        return self

    def parse_file(self, filename: str) -> PropertyParser:
        with open(filename, "r") as propFile:
            data = propFile.read()
        return self.read_string(data)

    def keys(self) -> KeysView[str]:
        return self.propDict.keys()

    def items(self) -> ItemsView[str, str]:
        return self.propDict.items()

    def write(self, fp):
        for key in self.propDict.keys():
            fp.write("%s=%s\n" % (key, self.propDict[key]))

    def has_option(self, name: str) -> bool:
        return name in self.propDict

    def get_option(self, name: str) -> Optional[str]:
        if self.has_option(name):
            return self.propDict[name]
        else:
            return None

    def set_option(self, name: str, value: str):
        self.propDict[name] = value

    def replace_option(self, name: str, value: str) -> bool:
        if self.has_option(name):
            if self.get_option(name) == value:
                return False
        self.set_option(name, value)
        return True

    def __eq__(self, other) -> bool:
        if isinstance(other, PropertyParser):
            return self.keys() == other.keys()
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Dict, Optional, Union

__all__ = ["PackageParser"]


def get_line_identation(line: str) -> int:
    return len(line) - (len(line.lstrip()))


def is_line_empty(line: str) -> bool:
    return len(line.strip()) == 0


class PackageParser(object):
    """"
    Parse the dumpsys output
    """
    def __init__(self, string: str):
        lines = list(map(lambda x: x.rstrip(), string.strip().splitlines()))
        self._data = PackageParser.PackageNode.parse(lines)

    @property
    def data(self):
        return self._data

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        if key in self._data:
            return self._data[key]
        else:
            return None

    """ Inner class """

    class PackageNode(object):
        def __init__(self, line: Optional[str], identation: int):
            self._data = dict()
            self._identation = identation
            if line:
                self._name = line.strip()
            else:
                self._name = None

        def __getitem__(self, key):
            return self._data[key]

        @property
        def identation(self):
            return self._identation

        def items(self):
            return self._data.items()

        def keys(self):
            return self._data.keys()

        def values(self):
            return self._data.values()

        @property
        def name(self):
            return self._name

        # noinspection PyUnresolvedReferences
        @staticmethod
        def parse(lines: List[str]) -> Dict[str, Union[dict, Dict[str, PackageNode]]]:
            result2 = dict()
            if len(lines) < 1:
                return result2

            cur_line_identation = -1
            cur_line = None
            nested_block = False

            for i in range(0, len(lines)):
                line: str = lines[i]
                if is_line_empty(line):
                    continue
                identation = get_line_identation(line)

                if i == 0 or identation == cur_line_identation:
                    line = line.strip()
                    # node = PackageParser.PackageNode(line, identation)
                    node = dict()
                    result2[line] = node
                    cur_line = line
                    cur_line_identation = identation
                    nested_block = False
                elif i > 0 and identation > cur_line_identation:
                    if not nested_block:
                        result2[cur_line] = PackageParser.PackageNode.parse(lines[i:])
                        nested_block = True
                else:
                    return result2
            return result2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Optional, List, Dict, Any

__all__ = ["Intent"]


class Intent(object):
    """
    See https://developer.android.com/studio/command-line/adb#IntentSpec
    """

    class Extras(object):
        def __init__(self):
            self.es: Dict[str, str] = dict()
            self.ez: Dict[str, bool] = dict()
            self.ei: Dict[str, int] = dict()
            self.el: Dict[str, int] = dict()
            self.ef: Dict[str, float] = dict()
            self.eu: Dict[str, str] = dict()
            self.ecn: Dict[str, str] = dict()
            self.eia: Dict[str, List[int]] = dict()
            self.ela: Dict[str, List[int]] = dict()
            self.efa: Dict[str, List[float]] = dict()
            self.grant_read_uri_permission = False
            self.grant_write_uri_permission = False
            self.exclude_stopped_packages = False
            self.include_stopped_packages = False
            self.raw_extras: Optional[str] = None

        def build(self) -> str:
            if self.raw_extras:
                return self.raw_extras

            args: List[str] = list()
            append_dict_argument(args, "--es", self.es)
            append_dict_argument(args, "--ez", self.ez)
            append_dict_argument(args, "--ei", self.ei)
            append_dict_argument(args, "--el", self.el)
            append_dict_argument(args, "--ef", self.ef)
            append_dict_argument(args, "--eu", self.eu)
            append_dict_argument(args, "--ecn", self.ecn)
            append_dict_argument(args, "--eia", self.eia)
            append_dict_argument(args, "--ela", self.ela)
            append_dict_argument(args, "--efa", self.efa)
            if self.grant_read_uri_permission:
                args.append("--grant-read-uri-permission")
            if self.grant_write_uri_permission:
                args.append("--grant-write-uri-permission")
            if self.exclude_stopped_packages:
                args.append("--exclude-stopped-packages")
            if self.include_stopped_packages:
                args.append("--include-stopped-packages")
            return " ".join(args)

    def __init__(self, action: str, extras: Optional[Extras] = Extras()):
        self.action = action
        self.data_uri: Optional[str] = None
        self.mime_type: Optional[str] = None
        self.category: Optional[str] = None
        self.component: Optional[str] = None
        self.flags: int = 0
        self.wait: bool = False
        self.user_id: int = 0
        self.extras = extras

    def waitforlaunchtocomplete(self, value: bool):
        self.wait = value

    def build(self) -> str:
        args = list()

        if self.wait:
            args.append("-W")

        if self.user_id > 0:
            args.append(f"--user {self.user_id}")

        args.append(f"-a {self.action}")
        append_str_argument(args, "-n", self.component)
        append_str_argument(args, "-d", self.data_uri)
        append_str_argument(args, "-t", self.mime_type)
        append_str_argument(args, "-c", self.category)
        append_int_argument(args, "-f", self.flags, 0)
        if self.extras:
            args.append(self.extras.build())
        return " ".join(args)


def append_dict_argument(target: List[str], key: str, option: Dict[str, Any]):
    if option:
        for k, v in option.items():
            string = f"{key} {k} "
            if isinstance(v, list):
                string += " ".join(str(x) for x in v)
            else:
                string += v
            target.append(string)


def append_str_argument(target: List[str], key: str, value: Optional[str]):
    if value:
        target.append(f"{key} {value}")


def append_int_argument(target: List[str], key: str, value: int, nullvalue: int):
    if value != nullvalue:
        target.append(f"{key} {value}")



"""
rm -fr out build dist src/pythonadb.egg-info
python3 setup.py sdist bdist_wheel
python3 -m twine upload dist/*
"""

import sys
import re
from setuptools import find_namespace_packages, setup

assert sys.version_info >= (3, 8, 0), "STB Utilities requires Python 3.8+"

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSIONFILE = "src/pythonadb/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
    name="pythonadb",
    version=verstr,
    author="Alessandro Crugnola",
    author_email="alessandro.crugnola@gmail.com",
    description="pure python adb client for mac/win/linux",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sephiroth74/python-adb-client",
    license="MIT",
    py_modules=[],
    package_dir={"": "src"},
    python_requires=">=3.8",
    packages=find_namespace_packages(where="src", exclude=("tests", "htmlconv")),
    test_suite="nose.collector",
    install_requires=[
        "verboselogs>=1.7",
        "zope.event>=4.5.0",
        "coloredlogs>=14.0",
        "optional.py>=1.1.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
